from pathlib import Path
from zipfile import ZipFile

# wheel platform tag: platform dynamic lib
PLATFORMS = {
    'win_amd64': 'cyber/lib/cyber.dll',
    'macosx_10_9_x86_64': 'cyber/lib/libcyber-arm64.dylib',
    'macosx_11_0_arm64': 'cyber/lib/libcyber.dylib',
    'manylinux_2_12_x86_64.manylinux2010_x86_64.musllinux_1_1_x86_64': 'cyber/lib/libcyber.so',
    # 'manylinux_2_12_arm64.manylinux2010_arm64.musllinux_1_1_arm64': 'cyber/lib/libcyber-arm64.so',
}

uni_wheels = Path('dist').glob('*py3-none-any.whl')

for uni_wheel_name in uni_wheels:
    for platform, lib in PLATFORMS.items():
        platform_wheel_name = uni_wheel_name.as_posix().replace('any', platform)

        zin = ZipFile(uni_wheel_name, 'r')
        zout = ZipFile(platform_wheel_name, 'w')

        for item in zin.infolist():
            buffer = zin.read(item.filename)
            if item.filename.startswith('cyber/lib/'):
                if item.filename == lib:
                    zout.writestr(item, buffer)
            else:
                zout.writestr(item, buffer)

        zout.close()
        zin.close()



from cyber import CyberVM, CyResultCode, CyType


# def test_validate_pass():
#     cyber = CyberVM()
#     cyber.validate('none')

#     assert cyber.last_result == CyResultCode.CY_Success

# def test_validate_token_error():
#     cyber = CyberVM()
#     cyber.validate('')

#     assert cyber.last_result == CyResultCode.CY_ErrorCompile

def test_validate_compile_error():
    cyber = CyberVM()
    cyber.validate('null')

    assert cyber.last_result == CyResultCode.CY_ErrorCompile



import re


class Writer:
    def __init__(self, out=None):
        self.items = []
        self.out = self.items.append
        if out:
            self.out = out
        self.indent = 4
        self.level = 0

    def __iadd__(self, other):
        self.line(other)
        return self

    def raw(self, string):
        self.out(string)

    def join(self, base: str):
        return base.join(self.items)

    def line(self, string=''):
        if string:
            self.out(' ' * self.indent * self.level + string + '\n')
        else:
            self.out('\n')

    def __enter__(self):
        self.level += 1
        return self

    def __exit__(self, *_):
        self.level -= 1


with open('src/cyber/src/cyber.h') as f:
    raw = f.read()

output = ''

thing = r'^.*[,;{]$'

matches = re.findall(thing, raw, flags=re.M)

definitions = []
chunk = []
in_chunk = False

for m in matches:
    if m.endswith('{'):
        in_chunk = True
    if in_chunk:
        chunk.append(m)

        if m.startswith('}'):
            in_chunk = False
            definitions.append('\n'.join(chunk))
            chunk = []

        continue
    definitions.append(m)

w = Writer()

w += '# GENERATED FILE DO NOT EDIT #'
w += ''
w += 'from ctypes import *'
w += 'from pathlib import Path'
w += 'import sys'
w += 'from enum import Enum'
w += 'import platform'
w += ''

# TODO: add linux-arm64 support
w += """
base_path = Path(__file__).parent / 'lib'

if sys.platform == 'win32':
    path = base_path / 'cyber.dll'
    lib = WinDLL(path.as_posix())
elif sys.platform == 'linux':
    path = base_path / 'libcyber.so'
    lib = CDLL(path.as_posix())
elif sys.platform == 'darwin':
    if platform.machine() == 'arm64':
        path = base_path / 'libcyber-arm64.dylib'
        lib = CDLL(path.as_posix())
    else:
        path = base_path / 'libcyber.dylib'
        lib = CDLL(path.as_posix())
"""
w += ''
w += '# GENERATED FILE DO NOT EDIT #'
w += ''

types = {
    'bool': 'c_bool',
    'double': 'c_double',
    'void*': 'c_void_p',
    'char*': 'c_char_p',
    'const char*': 'c_char_p',
    'size_t': 'c_size_t',
    'uint8_t': 'c_uint8',
    'uint64_t': 'c_uint64',
    'uint32_t': 'c_uint32',
    'int': 'c_int',
    'CyResultCode': 'c_int',
}


def typedef_int(w, d):
    m = re.match(r'typedef (\w+) (\w*)', d)
    t = m[1]
    types[f'{m[2]}*'] = f'POINTER({m[2]})'
    w += f'class {m[2]}({types.get(t, t)}):'
    with w:
        w += '...'


def typedef_struct(w, d):
    m = re.match(r'typedef struct (\w+) ({.*})?', d.replace('\n', ''))
    struct_name = m[1]
    types[f'{struct_name}*'] = f'POINTER({struct_name})'
    fields = []
    if m[2]:
        for f in m[2][1:-1].split(';'):
            if f:
                parts = f.strip().split(' ')
                name = parts[-1]
                c_type = ' '.join(parts[:-1])
                py_type = types.get(c_type, c_type)
                fields.append(f"('{name}', {py_type})")

    w += f'class {struct_name}(Structure):'
    with w:
        if fields:
            w += f"_fields_ = [{', '.join(fields)}]"
        else:
            w += '...'


def typedef_enum(w, d):
    m = re.match(r'typedef enum ({.*})? (\w+);', d.replace('\n', ''))
    values = m[1]
    enum_name = m[2]
    items = [v.strip().replace(' = 0', '') for v in values[1:-1].split(',') if v]

    w += f'class {enum_name}(Enum):'
    with w:
        for i, item in enumerate(items):
            w += f'{item} = {i}'


def typedef_funcptr_def(w, d):
    m = re.match(r'typedef (\w+) \(\*(\w+)\)(.*);', d)
    type_name = m[2]
    args = m[3][1:-1].split(', ')
    arg_types = [a.split(' ')[0] for a in args]
    argtypes = [types.get(m[1], m[1])]
    for arg in arg_types:
        argtypes.append(types.get(arg, arg))

    w += f'# {d}'
    w += f'{type_name} = CFUNCTYPE({", ".join(argtypes)})'


def func_def(w, d):
    m = re.match(r'(\w+\*?) (\w+)(.*);', d)
    func_name = m[2]
    ret = m[1]
    args = m[3][1:-1].split(', ')
    arg_types = [a.split(' ')[0] for a in args]
    argtypes = []
    for arg in arg_types:
        argtypes.append(types.get(arg, arg))

    w += f'# {d}'
    w += f'{func_name} = lib.{func_name}'
    if ret != 'void':
        w += f'{func_name}.restype = {types.get(ret, ret)}'
    w += f'{func_name}.argtypes = [{", ".join(argtypes)}]'


for d in definitions:
    if d.startswith('//'):
        continue
    if 'typedef' in d:
        if '(*' in d:
            typedef_funcptr_def(w, d)
        elif 'int' in d:
            typedef_int(w, d)
        elif 'struct' in d:
            typedef_struct(w, d)
        elif 'enum' in d:
            typedef_enum(w, d)
    else:
        if '=' in d:
            continue
        func_def(w, d)
    w += ''

with open('cyber/lib.py', 'w') as f:
    f.write(w.join(''))



from cyber import CyberVM, CyResultCode, CyType, CyValue


def test_python_function_specified_argtype(capsys):
    cyber = CyberVM()

    @cyber.function('core.test')
    def _new_func(one: str, two: int, three: float, four: bool):
        print(one, two, three, four)

    cyber.eval("test 'one' 2 3 4")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'one 2 3.0 True\n'


def test_python_function_specified_raw_cyvalue():
    cyber = CyberVM()

    output = []

    @cyber.function('core.test')
    def _new_func(value: CyValue):
        output.append(value)

    cyber.eval("test 'anything'")

    assert cyber.last_result == CyResultCode.CY_Success
    assert type(output[0]) == CyValue


def test_python_function_unspecified_argtypes(capsys):
    cyber = CyberVM()

    @cyber.function('core.test')
    def _new_func(one, two, three):
        print(one, two, three)

    cyber.eval("test 'one' 2 3")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'one 2.0 3.0\n'


def test_python_function_return_type():
    cyber = CyberVM()

    @cyber.function('core.test')
    def _test_func() -> str:
        return 'test'

    output = cyber.eval('test()')

    assert cyber.last_result == CyResultCode.CY_Success

    # TODO: I think cyber might not support this
    # assert output == 'test'



from pathlib import Path
from urllib.request import urlopen


BASE_URL = 'https://github.com/DaelonSuzuka/cyber/releases/download/latest/'


lib_urls = [
    'libcyber.so',
    # 'libcyber.so',
    'libcyber.dylib',
    'libcyber-arm64.dylib',
    'cyber.dll',
]


dst_dir = Path('cyber/lib')


for path in lib_urls:
    url = BASE_URL + path
    r = urlopen(url)
    
    with open(dst_dir / path, 'wb') as f:
        f.write(r.read())



# GENERATED FILE DO NOT EDIT #

from ctypes import *
from pathlib import Path
import sys
from enum import Enum
import platform


base_path = Path(__file__).parent / 'lib'

if sys.platform == 'win32':
    path = base_path / 'cyber.dll'
    lib = WinDLL(path.as_posix())
elif sys.platform == 'linux':
    path = base_path / 'libcyber.so'
    lib = CDLL(path.as_posix())
elif sys.platform == 'darwin':
    if platform.machine() == 'arm64':
        path = base_path / 'libcyber-arm64.dylib'
        lib = CDLL(path.as_posix())
    else:
        path = base_path / 'libcyber.dylib'
        lib = CDLL(path.as_posix())


# GENERATED FILE DO NOT EDIT #

class CyVM(Structure):
    ...

class CyModule(Structure):
    ...

class CyValue(c_uint64):
    ...

class CyResultCode(Enum):
    CY_Success = 0
    CY_ErrorToken = 1
    CY_ErrorParse = 2
    CY_ErrorCompile = 3
    CY_ErrorPanic = 4
    CY_ErrorUnknown = 5

class CyType(Enum):
    CY_TypeNone = 0
    CY_TypeBoolean = 1
    CY_TypeError = 2
    CY_TypeStaticAstring = 3
    CY_TypeStaticUstring = 4
    CY_TypeUserTag = 5
    CY_TypeUserTagLiteral = 6
    CY_TypeInteger = 7
    CY_TypeNumber = 8
    CY_TypeList = 9
    CY_TypeListIter = 10
    CY_TypeMap = 11
    CY_TypeMapIter = 12
    CY_TypeClosure = 13
    CY_TypeLambda = 14
    CY_TypeAstring = 15
    CY_TypeUstring = 16
    CY_TypeStringSlice = 17
    CY_TypeRawString = 18
    CY_TypeRawStringSlice = 19
    CY_TypeFiber = 20
    CY_TypeBox = 21
    CY_TypeNativeFunc1 = 22
    CY_TypeTccState = 23
    CY_TypeOpaquePtr = 24
    CY_TypeFile = 25
    CY_TypeDir = 26
    CY_TypeDirIter = 27
    CY_TypeSymbol = 28

class CStr(Structure):
    _fields_ = [('charz', c_char_p), ('len', c_size_t)]

class CyTypeId(c_uint32):
    ...

# typedef CyValue (*CyFunc)(CyVM* vm, CyValue* args, uint8_t nargs);
CyFunc = CFUNCTYPE(CyValue, POINTER(CyVM), POINTER(CyValue), c_uint8)

# typedef bool (*CyLoadModuleFunc)(CyVM* vm, CyModule* mod);
CyLoadModuleFunc = CFUNCTYPE(c_bool, POINTER(CyVM), POINTER(CyModule))

# CStr cyGetFullVersion();
cyGetFullVersion = lib.cyGetFullVersion
cyGetFullVersion.restype = CStr
cyGetFullVersion.argtypes = []

# CStr cyGetVersion();
cyGetVersion = lib.cyGetVersion
cyGetVersion.restype = CStr
cyGetVersion.argtypes = []

# CStr cyGetBuild();
cyGetBuild = lib.cyGetBuild
cyGetBuild.restype = CStr
cyGetBuild.argtypes = []

# CStr cyGetCommit();
cyGetCommit = lib.cyGetCommit
cyGetCommit.restype = CStr
cyGetCommit.argtypes = []

# CyVM* cyVmCreate();
cyVmCreate = lib.cyVmCreate
cyVmCreate.restype = POINTER(CyVM)
cyVmCreate.argtypes = []

# void cyVmDestroy(CyVM* vm);
cyVmDestroy = lib.cyVmDestroy
cyVmDestroy.argtypes = [POINTER(CyVM)]

# CyResultCode cyVmEval(CyVM* vm, CStr src, CyValue* outVal);
cyVmEval = lib.cyVmEval
cyVmEval.restype = c_int
cyVmEval.argtypes = [POINTER(CyVM), CStr, POINTER(CyValue)]

# CyResultCode cyVmValidate(CyVM* vm, CStr src);
cyVmValidate = lib.cyVmValidate
cyVmValidate.restype = c_int
cyVmValidate.argtypes = [POINTER(CyVM), CStr]

# CStr cyVmGetLastErrorReport(CyVM* vm);
cyVmGetLastErrorReport = lib.cyVmGetLastErrorReport
cyVmGetLastErrorReport.restype = CStr
cyVmGetLastErrorReport.argtypes = [POINTER(CyVM)]

# void cyVmRelease(CyVM* vm, CyValue val);
cyVmRelease = lib.cyVmRelease
cyVmRelease.argtypes = [POINTER(CyVM), CyValue]

# void cyVmRetain(CyVM* vm, CyValue val);
cyVmRetain = lib.cyVmRetain
cyVmRetain.argtypes = [POINTER(CyVM), CyValue]

# void* cyVmGetUserData(CyVM* vm);
cyVmGetUserData = lib.cyVmGetUserData
cyVmGetUserData.restype = c_void_p
cyVmGetUserData.argtypes = [POINTER(CyVM)]

# void cyVmSetUserData(CyVM* vm, void* userData);
cyVmSetUserData = lib.cyVmSetUserData
cyVmSetUserData.argtypes = [POINTER(CyVM), c_void_p]

# void cyVmAddModuleLoader(CyVM* vm, CStr name, CyLoadModuleFunc func);
cyVmAddModuleLoader = lib.cyVmAddModuleLoader
cyVmAddModuleLoader.argtypes = [POINTER(CyVM), CStr, CyLoadModuleFunc]

# void cyVmSetModuleFunc(CyVM* vm, CyModule* mod, CStr name, uint32_t numParams, CyFunc func);
cyVmSetModuleFunc = lib.cyVmSetModuleFunc
cyVmSetModuleFunc.argtypes = [POINTER(CyVM), POINTER(CyModule), CStr, c_uint32, CyFunc]

# void cyVmSetModuleVar(CyVM* vm, CyModule* mod, CStr name, CyValue val);
cyVmSetModuleVar = lib.cyVmSetModuleVar
cyVmSetModuleVar.argtypes = [POINTER(CyVM), POINTER(CyModule), CStr, CyValue]

# void* cyVmAlloc(CyVM* vm, size_t size);
cyVmAlloc = lib.cyVmAlloc
cyVmAlloc.restype = c_void_p
cyVmAlloc.argtypes = [POINTER(CyVM), c_size_t]

# void cyVmFree(CyVM* vm, void* ptr, size_t len);
cyVmFree = lib.cyVmFree
cyVmFree.argtypes = [POINTER(CyVM), c_void_p, c_size_t]

# CyValue cyValueNone();
cyValueNone = lib.cyValueNone
cyValueNone.restype = CyValue
cyValueNone.argtypes = []

# CyValue cyValueTrue();
cyValueTrue = lib.cyValueTrue
cyValueTrue.restype = CyValue
cyValueTrue.argtypes = []

# CyValue cyValueFalse();
cyValueFalse = lib.cyValueFalse
cyValueFalse.restype = CyValue
cyValueFalse.argtypes = []

# CyValue cyValueNumber(double n);
cyValueNumber = lib.cyValueNumber
cyValueNumber.restype = CyValue
cyValueNumber.argtypes = [c_double]

# CyValue cyValueInteger(int n);
cyValueInteger = lib.cyValueInteger
cyValueInteger.restype = CyValue
cyValueInteger.argtypes = [c_int]

# CyValue cyValueGetOrAllocStringInfer(CyVM* vm, CStr str);
cyValueGetOrAllocStringInfer = lib.cyValueGetOrAllocStringInfer
cyValueGetOrAllocStringInfer.restype = CyValue
cyValueGetOrAllocStringInfer.argtypes = [POINTER(CyVM), CStr]

# CyValue cyValueGetOrAllocAstring(CyVM* vm, CStr str);
cyValueGetOrAllocAstring = lib.cyValueGetOrAllocAstring
cyValueGetOrAllocAstring.restype = CyValue
cyValueGetOrAllocAstring.argtypes = [POINTER(CyVM), CStr]

# CyValue cyValueGetOrAllocUstring(CyVM* vm, CStr str, uint32_t charLen);
cyValueGetOrAllocUstring = lib.cyValueGetOrAllocUstring
cyValueGetOrAllocUstring.restype = CyValue
cyValueGetOrAllocUstring.argtypes = [POINTER(CyVM), CStr, c_uint32]

# CyValue cyValueAllocList(CyVM* vm);
cyValueAllocList = lib.cyValueAllocList
cyValueAllocList.restype = CyValue
cyValueAllocList.argtypes = [POINTER(CyVM)]

# CyValue cyValueAllocMap(CyVM* vm);
cyValueAllocMap = lib.cyValueAllocMap
cyValueAllocMap.restype = CyValue
cyValueAllocMap.argtypes = [POINTER(CyVM)]

# CyValue cyValueAllocNativeFunc(CyVM* vm, CyFunc func, uint32_t numParams);
cyValueAllocNativeFunc = lib.cyValueAllocNativeFunc
cyValueAllocNativeFunc.restype = CyValue
cyValueAllocNativeFunc.argtypes = [POINTER(CyVM), CyFunc, c_uint32]

# CyValue cyValueAllocOpaquePtr(CyVM* vm, void* ptr);
cyValueAllocOpaquePtr = lib.cyValueAllocOpaquePtr
cyValueAllocOpaquePtr.restype = CyValue
cyValueAllocOpaquePtr.argtypes = [POINTER(CyVM), c_void_p]

# CyValue cyValueTagLiteral(CyVM* vm, CStr str);
cyValueTagLiteral = lib.cyValueTagLiteral
cyValueTagLiteral.restype = CyValue
cyValueTagLiteral.argtypes = [POINTER(CyVM), CStr]

# CyTypeId cyValueGetTypeId(CyValue val);
cyValueGetTypeId = lib.cyValueGetTypeId
cyValueGetTypeId.restype = CyTypeId
cyValueGetTypeId.argtypes = [CyValue]

# double cyValueAsNumber(CyValue val);
cyValueAsNumber = lib.cyValueAsNumber
cyValueAsNumber.restype = c_double
cyValueAsNumber.argtypes = [CyValue]

# bool cyValueToBool(CyValue val);
cyValueToBool = lib.cyValueToBool
cyValueToBool.restype = c_bool
cyValueToBool.argtypes = [CyValue]

# bool cyValueAsBool(CyValue val);
cyValueAsBool = lib.cyValueAsBool
cyValueAsBool.restype = c_bool
cyValueAsBool.argtypes = [CyValue]

# int cyValueAsInteger(CyValue val);
cyValueAsInteger = lib.cyValueAsInteger
cyValueAsInteger.restype = c_int
cyValueAsInteger.argtypes = [CyValue]

# uint32_t cyValueAsTagLiteralId(CyValue val);
cyValueAsTagLiteralId = lib.cyValueAsTagLiteralId
cyValueAsTagLiteralId.restype = c_uint32
cyValueAsTagLiteralId.argtypes = [CyValue]

# CStr cyValueToTempString(CyVM* vm, CyValue val);
cyValueToTempString = lib.cyValueToTempString
cyValueToTempString.restype = CStr
cyValueToTempString.argtypes = [POINTER(CyVM), CyValue]

# CStr cyValueToTempRawString(CyVM* vm, CyValue val);
cyValueToTempRawString = lib.cyValueToTempRawString
cyValueToTempRawString.restype = CStr
cyValueToTempRawString.argtypes = [POINTER(CyVM), CyValue]

# size_t cyListLen(CyValue list);
cyListLen = lib.cyListLen
cyListLen.restype = c_size_t
cyListLen.argtypes = [CyValue]

# size_t cyListCap(CyValue list);
cyListCap = lib.cyListCap
cyListCap.restype = c_size_t
cyListCap.argtypes = [CyValue]

# CyValue cyListGet(CyVM* vm, CyValue list, size_t idx);
cyListGet = lib.cyListGet
cyListGet.restype = CyValue
cyListGet.argtypes = [POINTER(CyVM), CyValue, c_size_t]

# void cyListSet(CyVM* vm, CyValue list, size_t idx, CyValue val);
cyListSet = lib.cyListSet
cyListSet.argtypes = [POINTER(CyVM), CyValue, c_size_t, CyValue]

# void cyListAppend(CyVM* vm, CyValue list, CyValue val);
cyListAppend = lib.cyListAppend
cyListAppend.argtypes = [POINTER(CyVM), CyValue, CyValue]

# void cyListInsert(CyVM* vm, CyValue list, size_t idx, CyValue val);
cyListInsert = lib.cyListInsert
cyListInsert.argtypes = [POINTER(CyVM), CyValue, c_size_t, CyValue]




from cyber import CyberVM, CyResultCode, CyType, CyValue


def test_eval_none():
    cyber = CyberVM()

    output = cyber.eval('none')

    assert output == None
    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeNone


def test_eval_bool():
    cyber = CyberVM()

    output = cyber.eval('true')

    assert output == True
    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeBoolean

    output = cyber.eval('false')

    assert output == False
    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeBoolean


def test_eval_int():
    cyber = CyberVM()


# TODO: parameterize me
def test_eval_number():
    cyber = CyberVM()

    output = cyber.eval('1')

    assert output == 1
    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeNumber

    output = cyber.eval('1.5')

    assert output == 1.5
    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeNumber


def test_eval_cyvalue():
    cyber = CyberVM()

    script = """
        func foo(a):
            pass
            
        foo
    """
    output = cyber.eval(script)

    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeLambda
    assert type(output) == CyValue


def test_eval_static_a_string():
    cyber = CyberVM()

    output = cyber.eval("'string'")

    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeStaticAstring
    assert output == 'string'


def test_eval_static_u_string():
    cyber = CyberVM()

    output = cyber.eval("'💕'")

    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeStaticUstring
    assert output == '💕'


def test_eval_a_string():
    cyber = CyberVM()
    script = """
    t = 'string'
    s = '{t} interpolation'
    s
    """
    output = cyber.eval(script)

    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeAstring
    assert output == 'string interpolation'


def test_eval_u_string():
    cyber = CyberVM()
    script = """
    t = '💕'
    s = '{t} interpolation'
    s
    """
    output = cyber.eval(script)

    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeUstring
    assert output == '💕 interpolation'


def test_eval_string_slice():
    cyber = CyberVM()

    script = """
    t = 'string'
    s = '{t}'
    s[..3]
    """
    output = cyber.eval(script)

    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeStringSlice
    assert output == 'str'


def test_eval_rawstring():
    cyber = CyberVM()

    output = cyber.eval("rawstring('rawstring')")

    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeRawString
    assert output == b'rawstring'


def test_eval_rawstring_slice():
    cyber = CyberVM()

    script = """
    t = rawstring('rawstring')
    t[3..]
    """
    output = cyber.eval(script)

    assert cyber.last_result == CyResultCode.CY_Success
    assert cyber.last_output_type == CyType.CY_TypeRawStringSlice
    assert output == b'string'



from .vm import *


try:
    from IPython.core.magic import Magics, magics_class, line_cell_magic

    @magics_class
    class CyberMagics(Magics):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._cyber = CyberVM()
            
            @self._cyber.function('core.print')
            def _print(string: str):
                print(string)

        @line_cell_magic
        def cyber(self, line, cell=None):
            if cell is None:
                self._cyber.eval(line)
            else:
                self._cyber.eval(cell)

    def load_ipython_extension(ipython):
        ipython.register_magics(CyberMagics)

except ImportError:
    pass



from cyber import CyberVM, CyResultCode, CyType


def test_override_function(capsys):
    cyber = CyberVM()

    @cyber.function('print')
    def _print(string: str):
        print('<python>', string)

    cyber.eval("print 'hello'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == '<python> hello\n'


def test_add_function(capsys):
    cyber = CyberVM()

    @cyber.function('new_func')
    def _new_func(string: str):
        print(string)

    cyber.eval("new_func 'hello'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'hello\n'


def test_add_function_naked_decorator(capsys):
    cyber = CyberVM()

    @cyber.function
    def new_func(string: str):
        print(string)

    cyber.eval("new_func 'hello'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'hello\n'


def test_add_function_two_args(capsys):
    cyber = CyberVM()

    @cyber.function('new_func')
    def _new_func(one: str, two: str):
        print(one, two)

    cyber.eval("new_func 'one' 'two'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'one two\n'


def test_add_func_to_new_module(capsys):
    cyber = CyberVM()

    @cyber.function('new_module.new_func')
    def _new_func(one: float, two: float):
        print(one, two)

    script = """
        import mod 'new_module'
        mod.new_func 1 2
    """
    cyber.eval(script)

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == '1.0 2.0\n'


def test_module_inheritance(capsys):
    cyber = CyberVM()

    class Module(cyber.module('new_module')):
        def test(self):
            print('test')
        def test2(self):
            print('test2')

    script = """
        import mod 'new_module'
        mod.test()
        mod.test2()
    """
    cyber.eval(script)

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'test\ntest2\n'


def test_module_decorator(capsys):
    cyber = CyberVM()

    @cyber.module('new_module')
    class Module:
        def test(self):
            print('test')
        def test2(self):
            print('test2')

    script = """
        import mod 'new_module'
        mod.test()
        mod.test2()
    """
    cyber.eval(script)

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'test\ntest2\n'


def test_variables(capsys):
    cyber = CyberVM()

    @cyber.function('print')
    def _print(string: str):
        print(string)

    cyber.variable('one', 1)
    cyber.variable('two', 'two')
    cyber.variable('mod.three', 3)
    cyber.variable('mod.foure', '4')

    script = """
        import mod 'mod'
        print one
        print two
        print mod.three
        print mod.foure
    """
    cyber.exec(script)

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == '1\ntwo\n3\n4\n'


def test_module_with_variables(capsys):
    cyber = CyberVM()

    @cyber.module
    class core:
        var_none = None
        var_true = True
        var_false = False
        var_string = 'abc'
        var_integer = 7
        var_number = 15.5

        def print(self, string):
            print(string)

    script = """
        print var_none
        print var_true
        print var_false
        print var_string
        print var_integer
        print var_number
    """
    cyber.exec(script)

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'None\nTrue\nFalse\nabc\n7\n15.5\n'



from cyber import CyberVM, CyResultCode, CyType, CyberTokenError
import pytest


def get_printing_vm():
    cyber = CyberVM()

    # have to override print so we can capture its output
    @cyber.function('core.print')
    def _print(string: str):
        print(string)

    return cyber


def test_string_eval(capsys):
    cyber = get_printing_vm()

    cyber.eval("print 'hello'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'hello\n'


def test_bytes_eval(capsys):
    cyber = get_printing_vm()

    cyber.eval(b"print 'hello'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'hello\n'


def test_multiple_eval(capsys):
    cyber = get_printing_vm()

    cyber.eval(b"print 'hello'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'hello\n'

    cyber.eval(b"print 'hello again'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'hello again\n'


def test_multiple_eval_after_failure(capsys):
    cyber = get_printing_vm()

    with pytest.raises(CyberTokenError):
        cyber.eval(b"print 'unterminated string")

    assert cyber.last_result == CyResultCode.CY_ErrorToken
    captured = capsys.readouterr() # clear stdout

    cyber.eval(b"print 'hello'")

    assert cyber.last_result == CyResultCode.CY_Success
    captured = capsys.readouterr()
    assert captured.out == 'hello\n'


from ctypes import *
import inspect
from .lib import *
from . import lib
from textwrap import dedent
import builtins


# *************************************************************************** #


class CyberTokenError(Exception):
    ...
class CyberParseError(Exception):
    ...
class CyberCompileError(Exception):
    ...
class CyberPanicError(Exception):
    ...
class CyberUnknownError(Exception):
    ...


def cstr(s) -> CStr:
    if isinstance(s, str):
        s = dedent(s).encode()
    return CStr(c_char_p(s), len(s))


def py_to_cyvalue(vm, value) -> CyValue:
    match type(value):
        case builtins.bool:
            return cyValueTrue() if value else cyValueFalse()
        case builtins.int:
            return cyValueInteger(value)
        case builtins.float:
            return cyValueNumber(value)
        case builtins.str:
            return cyValueGetOrAllocStringInfer(vm, cstr(value))
        case _:
            return cyValueNone()


def cyvalue_to_py(vm, cyvalue):
    match CyType(cyValueGetTypeId(cyvalue).value):
        case CyType.CY_TypeNone:
            return None
        case CyType.CY_TypeBoolean:
            return cyValueAsBool(cyvalue)
        case CyType.CY_TypeInteger:
            return cyValueAsInteger(cyvalue)
        case CyType.CY_TypeNumber:
            return cyValueAsNumber(cyvalue)
        case CyType.CY_TypeStaticAstring:
            return cyValueToTempString(vm, cyvalue).charz.decode()
        case CyType.CY_TypeStaticUstring:
            return cyValueToTempString(vm, cyvalue).charz.decode()
        case CyType.CY_TypeAstring:
            return cyValueToTempString(vm, cyvalue).charz.decode()
        case CyType.CY_TypeUstring:
            return cyValueToTempString(vm, cyvalue).charz.decode()
        case CyType.CY_TypeStringSlice:
            return cyValueToTempString(vm, cyvalue).charz.decode()
        case CyType.CY_TypeRawString:
            return cyValueToTempRawString(vm, cyvalue).charz
        case CyType.CY_TypeRawStringSlice:
            return cyValueToTempRawString(vm, cyvalue).charz
        case _:
            return cyvalue


def generate_callback_wrapper(func):
    sig = inspect.signature(func)
    nargs = len(sig.parameters)
    return_type = sig.return_annotation

    # TODO: probably not a robust way to do this
    if 'self' in sig.parameters:
        nargs = nargs - 1

    @CyFunc
    def wrapper(vm, args, nargs):
        # convert args based on registered function signature
        _args = []
        i = 0
        for name, param in sig.parameters.items():
            # TODO: probably not a robust way to do this
            if name == 'self':
                _args.append(0)
                continue
            match param.annotation:
                case builtins.str:
                    _args.append(cyValueToTempString(vm, args[i]).charz.decode())
                case builtins.int:
                    _args.append(int(cyValueAsNumber(args[i])))
                case builtins.float:
                    _args.append(cyValueAsNumber(args[i]))
                case builtins.bool:
                    _args.append(cyValueToBool(args[i]))
                case lib.CyValue: # raw cyValue
                    _args.append(args[i])
                case _: # no type specified, try to auto-convert
                    _args.append(cyvalue_to_py(vm, args[i]))
            i += 1

        raw_ret = func(*_args)

        ret = 0

        # TODO: FFI return types
        # convert return type
        # match return_type:
            # case None:
            #     ret = cyValueNone()
            # case builtins.str:
            #     ret = cyValueGetOrAllocStringInfer(vm, cstr(raw_ret))

        return ret

    return wrapper, nargs


class Module:
    def __init__(self, cyber, name, contents) -> None:
        self.cyber = cyber
        self.name = name
        self.functions = contents['funcs']
        self.variables = contents['vars']
        self.build()

    def build(self):
        @CyLoadModuleFunc
        def load_module(vm, mod):
            # print(f'[ContextModule]: loading module: {self.name}')
            for func_info in self.functions:
                # print(f'[ContextModule]: registering function: {name}')
                self.cyber.set_module_func(mod, *func_info)

            for var_info in self.variables:
                self.cyber.set_module_var(mod, *var_info)

            return True

        self.load_module = load_module
        self.cyber.add_module_loader(self.name, load_module)


class CyberVM:
    def __init__(self) -> None:
        self.last_result = None
        self.last_output_type = None
        self.last_output = None

        self.vm = cyVmCreate()
        self.modules = []
        self.pending_modules = {}
        self.pending_module_classes = []

    def add_module_loader(self, name, loader):
        cyVmAddModuleLoader(self.vm, cstr(name), loader)

    def set_module_func(self, mod, name, nargs, func):
        cyVmSetModuleFunc(self.vm, mod, cstr(name), nargs, func)

    def set_module_var(self, mod, name, value):
        cyVmSetModuleVar(self.vm, mod, cstr(name), py_to_cyvalue(self.vm, value))

    def _ensure_module(self, module_name):
        if module_name not in self.pending_modules:
            self.pending_modules[module_name] = {'funcs':[], 'vars':[]}
    
    def function(self, name):
        if isinstance(name, str):
            module_name = 'core'
            func_name = name
            if '.' in name:
                module_name, func_name = name.split('.')
            def _decorator(func):
                wrapper, nargs = generate_callback_wrapper(func)            
                self._ensure_module(module_name)
                self.pending_modules[module_name]['funcs'].append((func_name, nargs, wrapper))
                return func
            return _decorator
        else:
            # we ARE the decorator
            func = name
            module_name = 'core'
            func_name = func.__name__
            wrapper, nargs = generate_callback_wrapper(func)            

            self._ensure_module(module_name)
            self.pending_modules[module_name]['funcs'].append((func_name, nargs, wrapper))

            return func
        
    def variable(self, name, value):
        module_name = 'core'
        var_name = name
        if '.' in name:
            module_name, var_name = name.split('.')

        self._ensure_module(module_name)
        self.pending_modules[module_name]['vars'].append((var_name, value))

    def module(self, name):
        if isinstance(name, str):
            class ModuleClass:
                _name = name

                def __new__(cls, kls):
                    """should only be called as a class decorator"""
                    kls._name = cls._name
                    self.pending_module_classes.append(kls)
                    return cls, kls
                    
            self.pending_module_classes.append(ModuleClass)
            return ModuleClass
        else:
            # we ARE the decorator
            kls = name
            kls._name = kls.__name__
            
            self.pending_module_classes.append(kls)
            return kls

    def build_pending_modules(self):
        def get_funcs(klass):
            for func in [m for m in dir(klass) if callable(getattr(klass, m)) and not m.startswith('__')]:
                dec = self.function(f'{klass._name}.{func}')
                dec(getattr(klass, func))

        def get_vars(klass):
            for var in [m for m in dir(klass) if not callable(getattr(klass, m)) and not m.startswith('__')]:
                self.variable(f'{klass._name}.{var}', getattr(klass, var))

        for klass in self.pending_module_classes:
            get_funcs(klass)
            get_vars(klass)
            for sub in klass.__subclasses__():
                get_funcs(sub)
                get_vars(sub)

        for module_name, contents in self.pending_modules.items():
            mod = Module(self, module_name, contents)
            self.modules.append(mod)

        self.pending_modules.clear()
        self.pending_module_classes.clear()

    def validate(self, src: str | bytes):
        self.build_pending_modules()

        self.last_output_type = None
        self.last_output = None
        result = cyVmValidate(self.vm, cstr(src))
        self.last_result = CyResultCode(result)
        return cyVmGetLastErrorReport(self.vm).charz
        # if self.last_result != CyResultCode.CY_Success:
        #     match self.last_result:
        #         case CyResultCode.CY_ErrorToken:
        #             raise CyberTokenError
        #         case CyResultCode.CY_ErrorParse:
        #             raise CyberParseError
        #         case CyResultCode.CY_ErrorCompile:
        #             raise CyberCompileError
        #         case CyResultCode.CY_ErrorPanic:
        #             raise CyberPanicError
        #         case CyResultCode.CY_ErrorUnknown:
        #             raise CyberUnknownError
        #     raise CyberUnknownError
        
    def eval(self, src: str | bytes):
        self.build_pending_modules()

        self.last_output_type = None
        self.last_output = None
        out = CyValue()
        result = cyVmEval(self.vm, cstr(src), pointer(out))
        self.last_result = CyResultCode(result)
        if self.last_result != CyResultCode.CY_Success:
            report = cyVmGetLastErrorReport(self.vm).charz
            match self.last_result:
                case CyResultCode.CY_ErrorToken:
                    raise CyberTokenError(report)
                case CyResultCode.CY_ErrorParse:
                    raise CyberParseError(report)
                case CyResultCode.CY_ErrorCompile:
                    raise CyberCompileError(report)
                case CyResultCode.CY_ErrorPanic:
                    raise CyberPanicError(report)
                case CyResultCode.CY_ErrorUnknown:
                    raise CyberUnknownError(report)
            raise CyberUnknownError
        
        self.last_output_type = CyType(cyValueGetTypeId(out).value)
        self.last_output = cyvalue_to_py(self.vm, out)
        return self.last_output

    def exec(self, src: str | bytes) -> None:
        self.eval(src)


from cyber import (
    CyberVM,
    CyResultCode,
    CyType,
    CyberTokenError,
    CyberParseError,
    CyberCompileError,
    CyberPanicError,
    CyberUnknownError,
)
import pytest


def test_token_error():
    cyber = CyberVM()

    with pytest.raises(CyberTokenError):
        cyber.exec("print 'unterminated string")

    assert cyber.last_result == CyResultCode.CY_ErrorToken


def test_parse_error():
    cyber = CyberVM()

    script = """
        var
        var a: 123
        var b: 234    
    """

    with pytest.raises(CyberParseError):
        cyber.exec(script)

    assert cyber.last_result == CyResultCode.CY_ErrorParse


def test_compile_error():
    cyber = CyberVM()

    script = """
        func foo(a Vec2):
            pass
    """
    with pytest.raises(CyberCompileError):
        cyber.exec(script)

    assert cyber.last_result == CyResultCode.CY_ErrorCompile


def test_panic_error():
    cyber = CyberVM()

    with pytest.raises(CyberPanicError):
        cyber.exec('panic(#danger)')

    assert cyber.last_result == CyResultCode.CY_ErrorPanic


# TODO: how do I trigger this?
# def test_unknown_error():
#     cyber = CyberVM()
#     script = """

#     """
#     with pytest.raises(CyberUnknownError):
#         cyber.exec(script)

#     assert cyber.last_result == CyResultCode.CY_ErrorUnknown
